import os
import pytest
from pyrobinhood import PyRobinhood, AuthProvider


AUTH_DATA = {
    'access_token': os.environ['RB_TOKEN'],
    'refresh_token': os.environ['RB_REFRESH_TOKEN'],
    'client_id': os.environ['RB_CLIENT_ID']
}

class TestAuthProvider(AuthProvider):
    def get_auth_data(self):
        return AUTH_DATA

    def update_auth_data(self, data):
        AUTH_DATA = data

rb = PyRobinhood(TestAuthProvider())

def test_get_instrument():
    # valid symbol
    res = rb.get_instrument('AAPL')
    assert 'id' in res
    assert 'symbol' in res
    assert res['symbol'] == 'AAPL'

    # invalid symbol
    with pytest.raises(Exception):
        rb.get_instrument('HAHAHA')


def test_get_accounts():
    res = rb.get_accounts()
    assert len(res) >= 1

def test_get_positions():
    res = rb.get_positions()
    for p in res:
        assert 'instrument' in p
        assert 'quantity' in p
        assert 'average_buy_price' in p

def test_get_options_positions():
    res = rb.get_options_positions()
    for p in res:
        assert 'id' in p
        assert 'chain' in p
        assert 'symbol' in p
        assert 'quantity' in p
        assert 'strategy' in p

## Access Token will be reset here

def test_refresh_token():
    old_token = rb.token
    rb.__refresh_token__()
    new_token = rb.token
    assert old_token != new_token

def test_get_options_chains():
    symbol = 'MSFT'
    instrument = rb.get_instrument(symbol)
    instrument_id = instrument['id']

    results = rb.get_options_chains(instrument_id)
    for res in results:
        assert 'symbol' in res
        assert 'expiration_dates' in res

    target_result = [res for res in results if res['symbol'] == symbol]
    assert len(target_result) == 1    

def test_get_options_instrument():
    symbol = 'QQQ'
    stock_instrument_id = rb.get_instrument(symbol)['id']
    chains = rb.get_options_chains(stock_instrument_id)
    chain = [res for res in chains if res['symbol'] == symbol][0]

    exp_date = chain['expiration_dates'][0]
    option_type = 'call'

    results = rb.get_options_instrument(chain['id'], exp_date, option_type)
    for res in results:
        assert 'chain_id' in res and res['chain_id'] == chain['id']
        assert 'chain_symbol' in res and res['chain_symbol'] == symbol
        assert 'type' in res and res['type'] == option_type
        assert 'tradability' in res and res['tradability'] == 'tradable'
        assert 'strike_price' in res
        assert 'id' in res

def __get_random_options(symbol, option_type):
    stock_instrument_id = rb.get_instrument(symbol)['id']
    chains = rb.get_options_chains(stock_instrument_id)
    chain = [res for res in chains if res['symbol'] == symbol][0]
    exp_date = chain['expiration_dates'][0]

    results = rb.get_options_instrument(chain['id'], exp_date, option_type)
    return results[0]['id'], exp_date

def test_get_options_detail():
    symbol = 'FB'
    option_type = 'call'
    options_instrument_id, exp_date = __get_random_options(symbol, option_type)
    res = rb.get_options_detail(options_instrument_id)
    assert 'chain_symbol' in res and res['chain_symbol'] == symbol
    assert 'expiration_date' in res and res['expiration_date'] == exp_date
    assert 'id' in res and res['id'] == options_instrument_id
    assert 'type' in res and res['type'] == option_type
    assert 'tradability' in res and res['tradability'] == 'tradable'

def test_get_options_marketdata():
    symbol = 'FB'
    option_type = 'call'
    options_instrument_id, _ = __get_random_options(symbol, option_type)
    res = rb.get_options_marketdata(options_instrument_id)
    assert 'ask_price' in res
    assert 'bid_price' in res
    assert 'volume' in res
    assert 'previous_close_price' in res



INSTRUMENT_ENDPOINT = 'https://api.robinhood.com/instruments/'

def get_instrument(request, symbol, timeout=15):
    res = request.get(INSTRUMENT_ENDPOINT, params={'symbol': symbol}, timeout=timeout)
    res.raise_for_status()
    res = res.json()

    return res['results'][0]


ACCOUNTS_ENDPOINT = 'https://api.robinhood.com/accounts/'
POSITION_ENDPOINT = 'https://api.robinhood.com/positions/'
OPTIONS_POSITION_ENDPOINT = 'https://api.robinhood.com/options/aggregate_positions/'


def get_accounts(request, timeout=15):
    res = request.get(ACCOUNTS_ENDPOINT, timeout=timeout)
    res.raise_for_status()
    res = res.json()

    return res['results']


def get_positions(request, nonzero, timeout=15):
    nonzero_str = 'true' if nonzero else 'false'
    res = request.get(POSITION_ENDPOINT, params={'nonzero': nonzero_str}, timeout=timeout)
    res.raise_for_status()
    res = res.json()

    return res['results']


def get_options_positions(request, nonzero, timeout=15):
    nonzero_str = 'true' if nonzero else 'false'
    res = request.get(
        OPTIONS_POSITION_ENDPOINT,
        params={'nonzero': nonzero_str},
        timeout=timeout)
    res.raise_for_status()
    res = res.json()

    return res['results']


import requests
from .auth_provider import AuthProvider

from .instrument import get_instrument 
from .account import get_accounts, get_positions, get_options_positions
from .options import get_options_chains, get_options_instrument, get_options_detail, get_options_marketdata
from .token import post_refresh_token

class PyRobinhood:

    def __init__(self, auth_provider):
        self.session = requests.session()
        self.headers = {
            "Accept": "*/*",
            "Accept-Encoding": "gzip, deflate",
            "Accept-Language": "en;q=1, fr;q=0.9, de;q=0.8, ja;q=0.7, nl;q=0.6, it;q=0.5",
            "X-Robinhood-API-Version": "1.0.0",
            "Connection": "keep-alive",
            "User-Agent": "Robinhood/823 (iPhone; iOS 7.1.2; Scale/2.00)"
        }
        self.session.headers = self.headers
        self.auth_provider = auth_provider

        self.token = auth_provider.get_auth_data()['access_token']
        self.refresh_token = auth_provider.get_auth_data()['refresh_token']

        self.__auth_failed = False

    def __non_auth_request__(self, method, *args, **kwargs):
        if 'authorization' in self.session.headers:
            del self.session.headers['authorization']
        return method(self.session, *args, **kwargs)

    def __refresh_token__(self):
        refresh_token = self.auth_provider.get_auth_data()['refresh_token']
        client_id = self.auth_provider.get_auth_data()['client_id']
        res = post_refresh_token(self.session, client_id, refresh_token)

        self.token = res['access_token']
        res['client_id'] = client_id
        self.auth_provider.update_auth_data(res)

    def __auth_request__(self, method, *args, **kwargs):
        try:
            self.session.headers.update({ 'authorization': 'Bearer ' + self.token })
            res = method(self.session, *args, **kwargs)
            self.__auth_failed = False
            return res
        except requests.exceptions.HTTPError as e:
            if e.response.status_code == 401 and not self.__auth_failed:
                self.__auth_failed = True
                self.__refresh_token__()
                return self.__auth_request__(method, *args, **kwargs)
            else:
                raise e

    ####### Instrument #######
    def get_instrument(self, symbol):
        return self.__non_auth_request__(get_instrument, symbol)

    ####### Account #######
    def get_accounts(self):
        return self.__auth_request__(get_accounts)

    def get_positions(self, nonzero=True):
        return self.__auth_request__(get_positions, nonzero)

    def get_options_positions(self, nonzero=True):
        return self.__auth_request__(get_options_positions, nonzero)

    ####### Options #######
    def get_options_chains(self, instrument_id):
        return self.__non_auth_request__(get_options_chains, instrument_id)

    def get_options_instrument(self, chain_id, exp_date, option_type, state='active', tradability='tradable'):
        return self.__non_auth_request__(get_options_instrument, chain_id, exp_date, option_type, state, tradability)

    def get_options_detail(self, options_instrument_id):
        return self.__non_auth_request__(get_options_detail, options_instrument_id)

    def get_options_marketdata(self, options_instrument_id):
        return self.__auth_request__(get_options_marketdata, options_instrument_id)


try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup
from setuptools import find_packages

setup(name='pyrobinhood',
      version='0.1',
      description='Python Robinhood client',
      author='Daniel Wang',
      author_email='danielwpz@gmail.com',
      license='MIT',
      packages=find_packages(exclude=['tests*']),
      install_requires=[
          'requests'
      ],
      zip_safe=False)



OAUTH_TOKEN_ENDPOINT = 'https://api.robinhood.com/oauth2/token/'


def post_refresh_token(request, client_id, refresh_token, timeout=15):
    res = request.post(
        OAUTH_TOKEN_ENDPOINT,
        json={
            'client_id': client_id,
            'refresh_token': refresh_token,
            "grant_type" : "refresh_token",
            "scope" : "web_limited"
        },
        timeout=timeout
    )
    res.raise_for_status()
    res = res.json()

    return res


OPTIONS_CHAIN_ENDPOINT = 'https://api.robinhood.com/options/chains/'
OPTIONS_INSTRUMENT_ENDPOINT = 'https://api.robinhood.com/options/instruments/'
OPTIONS_DETAIL_ENDPOINT = 'https://api.robinhood.com/options/instruments/%s/'
OPTIONS_MARKETDATA_ENDPOINT = 'https://api.robinhood.com/marketdata/options/%s/'


def get_options_chains(request, instrument_id, timeout=15):
    res = request.get(
        OPTIONS_CHAIN_ENDPOINT, 
        params={'equity_instrument_ids': instrument_id}, 
        timeout=timeout)
    res.raise_for_status()
    res = res.json()

    return res['results']


def get_options_instrument(request, chain_id, expiration_date, option_type, state, tradability, timeout=15):
    res = request.get(
        OPTIONS_INSTRUMENT_ENDPOINT, 
        params={
            'chain_id': chain_id,
            'expiration_dates': expiration_date,
            'type': option_type,
            'state': state,
            'tradability': tradability
        }, 
        timeout=timeout)
    res.raise_for_status()
    res = res.json()
    return res['results']


def get_options_detail(request, options_instrument_id, timeout=15):
    endpoint = OPTIONS_DETAIL_ENDPOINT % options_instrument_id
    res = request.get(endpoint, timeout=timeout)
    res.raise_for_status()
    res = res.json()
    return res


def get_options_marketdata(request, options_instrument_id, timeout=15):
    endpoint = OPTIONS_MARKETDATA_ENDPOINT % options_instrument_id
    res = request.get(endpoint, timeout=timeout)
    res.raise_for_status()
    res = res.json()
    return res